# For General purpose functions
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from tqdm.auto import tqdm

# For PCA
from sklearn.decomposition import PCA

# For distance matrix calculation and clustering
from scipy.spatial.distance import pdist, squareform
from sklearn.cluster import KMeans
from tslearn.metrics import dtw

# For latent space visualization
from sklearn.manifold import MDS


# Define key parameters
seed = 0
np.random.seed(seed)


class trajectory:

    def __init__(self, simulation_df, states_df=None):        
        """
        Initializes the TrajectoryClustering object.

        Parameters:
        simulation_df (pd.DataFrame): DataFrame containing the simulation data.
        states_df (pd.DataFrame, optional): DataFrame containing the states data. Default is None.
        """
        self.simulation_df = simulation_df
        self.node_list = list(simulation_df.columns.drop(['model_id','timepoint']))

        # Only assign states_df if it is not None
        if states_df is not None:
            self.states_df = states_df

    ##### In this part of the script, we focus on visualizing the whole simulation trajectory by performing PCA on the simulation data. ######
    def pca_trajectory(self, n_components = 10):
        """
        Perform PCA on the given simulation DataFrame and return the transformed DataFrame.

        Parameters:
        simulation_df (pd.DataFrame): The input DataFrame containing simulation data. 
                                        It must include 'model_id' and 'timepoint' columns.
        n_components (int): The number of principal components to keep. Default is 10.

        Returns:
        pd.DataFrame: A DataFrame containing the principal components, along with the 
                        'model_id' and 'timepoint' columns from the original DataFrame.
        """
        simulation_df = self.simulation_df

        # Initialize PCA (let's reduce to 2 principal components for this example)
        pca = PCA(n_components=n_components)

        # Fit and transform the data
        df_pca = simulation_df.drop(['model_id','timepoint'], axis = 1) 
        pca_result = pca.fit_transform(df_pca)

        # Convert the result back to a DataFrame for easier interpretation
        pca_df = pd.DataFrame(data=pca_result, index=df_pca.index)

        # number pca column
        number_list = list(range(pca_result.shape[1]))
        str_list = [str(i+1) for i in number_list]
        pca_df.columns = ['pc' + s for s in str_list]

        # Add model_id and timepoint backinto dataframe
        pca_df['model_id'] = simulation_df['model_id']
        pca_df['timepoint'] = simulation_df['timepoint'].astype('float')

        # Store the PCA DataFrame
        self.pca_df = pca_df

    ##### In this part of the script, we calculate the distance matrix, based on either endpoint simulation or whole trajectory ######
    def calculate_distancematrix(self, 
                                 mode = ['endpoint', 'trajectory'],
                                 timepoint = None):
        
        # Extract the simulation DataFrame
        simulation_df = self.simulation_df

        # Check if the mode is valid
        if mode not in ['endpoint', 'trajectory']:
            raise ValueError("Mode must be either 'endpoint' or 'trajectory'.")
        
        # Calculate the distance matrix using endpoint
        if mode == 'endpoint':
            print("Calculating distance matrix for endpoint simulation...")
            # If mode is 'endpoint', we only consider the last timepoint
            if timepoint is None:
                timepoint = simulation_df['timepoint'].max()
            simulation_df = simulation_df.loc[simulation_df['timepoint'] == timepoint].set_index('model_id')
            model_name = simulation_df.index
            simulation_df = simulation_df.drop(columns=['timepoint'])
            simulation_array = simulation_df.to_numpy()
            distance_matrix = squareform(pdist(simulation_array, metric='euclidean'))
            distance_matrix = pd.DataFrame(distance_matrix, columns=model_name, index=model_name)

        # Calculate the distance matrix using whole trajectory
        elif mode == 'trajectory':
            print("Calculating distance matrix for whole trajectory...")
            # Preparing the calculating distance
            simulation_df.model_id = simulation_df.model_id.astype('category')
            node_list = simulation_df.columns.drop(['timepoint','model_id'])
            model_name = simulation_df.model_id.unique()
            model_original_all = {}
            
            # Loop through each model and extract the trajectory data
            for i in model_name:
                model_original = simulation_df.loc[simulation_df.model_id == i, node_list].values
                model_original_all[i] = np.array(model_original)
            pca_all_trajectory = np.array(list(model_original_all.values()))
            
            # Calculate the distance matrix using DTW
            num_trajectories = len(pca_all_trajectory)
            distance_matrix = np.zeros((num_trajectories, num_trajectories))
            for i in tqdm(range(num_trajectories)):
                for j in range(num_trajectories):
                    distance_matrix[i, j] = dtw(pca_all_trajectory[i], pca_all_trajectory[j])
            
            # Convert the distance matrix to a DataFrame
            distance_matrix = pd.DataFrame(distance_matrix, index=model_name, columns=model_name)
        
        # Return the distance matrix
        self.distance_matrix = distance_matrix
        print("Distance matrix calculated successfully.")
        
    # This script needs to be optimized more....    
    # def optimize_cluster(self,data = 'pca', n_cluster = 15, method = 'euclidean'):
    #     # Setup the variables
    #     pca_df = self.pca_df
    #     pca_df.model_id = pca_df.model_id.astype('category')
    #     model_name = pca_df.model_id.cat.categories

    #     simulation_df = self.simulation_df
    #     simulation_df.model_id = simulation_df.model_id.astype('category')
        
    #     model_name = pca_df.model_id.cat.categories
    #     model_pca_all = {}

    #     node_list = self.node_list

    #     if data == 'pca':
    #         model_pca_all = {}
    #         for i in model_name:
    #             model_pca = pca_df.loc[pca_df.model_id == i,['pc1','pc2']].values
    #             model_pca_all[i] = np.array(model_pca)
    #         pca_all_trajectory = np.array(list(model_pca_all.values()))

    #     elif data == 'original':
    #         model_original_all = {}
    #         for i in model_name:
    #             model_original = simulation_df.loc[simulation_df.model_id == i,node_list].values
    #             model_original_all[i] = np.array(model_original)
    #         pca_all_trajectory = np.array(list(model_original_all.values()))   

    #     # Calculate the optimal number of clusters
    #     distortions = []
    #     K = range(1, n_cluster)

    #     # For loop to calculate inertia for each k
    #     for k in tqdm(K):
    #         tsmodel = TimeSeriesKMeans(n_clusters=k, metric= method, random_state=0, verbose = False)
    #         tsmodel.fit(pca_all_trajectory)
    #         distortions.append(tsmodel.inertia_)

    #     plt.plot(K, distortions, 'bx-')
    #     plt.xlabel('k')
    #     plt.ylabel('Inertia')
    #     plt.title('Elbow Method For Optimal k')
    #     plt.show()
    
    def calculate_kmean_cluster(self, 
                                n_cluster, 
                                random_state = 12345):

        # Get the distance matrix
        distance_matrix = self.distance_matrix

        # Euclidean k-means
        kmeans = KMeans(n_clusters= n_cluster, random_state=random_state)
        kmeans.fit(distance_matrix)
        clusters = kmeans.labels_
        cluster_dict = dict(zip(list(distance_matrix.index),list(clusters)))

        # Store the updated DataFrame and cluster dictionary
        self.cluster_dict = cluster_dict

        print(f"Calculated k-means clustering with {n_cluster} clusters.")
    
    def calculate_MDS(self, random_state = 12345):
        
        # Calculate the MDS coordinates
        distance_matrix = self.distance_matrix
        mds = MDS(dissimilarity='precomputed', random_state=random_state, n_init=4)
        coords = mds.fit_transform(distance_matrix)

        # Store the MDS coordinates in a DataFrame
        coords_vis = pd.DataFrame(coords, columns = ['x', 'y'], index = distance_matrix.index)
        self.mds_coords = coords_vis

    # All the plotting functions are here 
    def plot_pca_trajectory(self, fig_size=(8,6), 
                    color='model_id', size=10,
                    plot_cluster = False, save_fig = False):
        pca_df = self.pca_df
        
        # Adjust figure size
        plt.figure(figsize = fig_size)
        
        # Line plot using seaborn
        plot = sns.lineplot(data = pca_df, 
                    x = 'pc1',y='pc2',
                    units = 'model_id', estimator = None, lw=2, alpha = .3,
                    sort = False, legend = False)

        if plot_cluster == False:
            # Line plot using seaborn
            plot2 = sns.lineplot(data = pca_df,
                            x = 'pc1', y = 'pc2',
                            units = 'model_id', estimator = None, 
                            hue = color, sort = False,
                            marker = 'o', markersize = size, 
                            linewidth = 2,
                            alpha = .5, legend = False 
                            )
        else:
            # Check if kmean_cluster is already calculated
            if not hasattr(self, 'cluster_dict'):
                print("k-mean cluster not calculated")
                return
            pca_df['kmean_cluster'] = pca_df['model_id'].map(self.cluster_dict)
            kmean_cluster = pca_df.groupby(['timepoint','kmean_cluster'])[['pc1','pc2']].mean()
            plot2 = sns.lineplot(data = kmean_cluster, 
                            x = 'pc1',y='pc2',
                            hue = 'kmean_cluster',
                            sort = False, marker = 'o', linewidth = 5, markersize = 10)
            
        # Show the plot
        plot.grid(True)
        if save_fig == True:
            plt.savefig('pca_trajectory.png', dpi=600, bbox_inches='tight')
            plt.savefig('pca_trajectory.pdf', bbox_inches='tight')
        plt.show()
        plt.close()

    def plot_model_distance_space(self, save_fig = False):
        distance_matrix = self.distance_matrix
        
        # Plot Euclidean distance using clustermap
        g = sns.clustermap(distance_matrix, figsize = (20,20))

        # Hiding the row dendrogram
        g.ax_row_dendrogram.set_visible(False)

        # Hiding the column dendrogram
        g.ax_col_dendrogram.set_visible(False)
        if save_fig == True:
            plt.savefig('distance_matrix.png', dpi=600, bbox_inches='tight')
            plt.savefig('distance_matrix.pdf', bbox_inches='tight')
        plt.show()
        plt.close()
    
    def plot_MDS(self, plot_cluster = False,
                 fig_size=(8, 6), alpha=0.5, s=50,
                 save_fig = False):

        coords = self.mds_coords
        plt.figure(figsize=fig_size)
        if plot_cluster:
            # Add the cluster labels to the coordinates
            coords['kmean_cluster'] = coords.index.map(self.cluster_dict)
            sns.scatterplot(x=coords['x'], y=coords['y'],
                            hue=coords['kmean_cluster'],
                            palette='tab10',
                            alpha=alpha, s=s)
            plt.legend(title="Clusters")
        else:
            # Plot without cluster labels
            sns.scatterplot(x=coords['x'], y=coords['y'],
                            alpha=alpha, s=s)
        if save_fig == True:
            plt.savefig('mds_projection.png', dpi=600, bbox_inches='tight')
            plt.savefig('mds_projection.pdf', bbox_inches='tight')
        plt.grid(False)
        plt.show()

    def plot_trajectory_variance(self, fig_size=(15, 7), save_fig = False):
        # Assuming `model_mtx` is your DataFrame with genes as columns and 'timepoint' as one of the columns
        model_mtx = self.simulation_df

        # Group by 'timepoint' and calculate variance for each gene
        variance_results = model_mtx.groupby('timepoint')[model_mtx.columns[:-2]].var()
        variance_results['avg'] = variance_results.mean(axis = 1)
        sns.clustermap(variance_results.drop(columns = ['avg']), 
                       row_cluster=False, cmap = 'viridis', 
                       figsize = fig_size)
        if save_fig == True:
            plt.savefig('trajectory_variance.png', dpi=600, bbox_inches='tight')
            plt.savefig('trajectory_variance.pdf', bbox_inches='tight')
        plt.show()
        plt.close()
    
    def plot_node_trajectory(self, 
                             node, 
                             fig_width = 4,
                             fig_height = 4,
                             n_timesteps = 20,
                             save_fig = False):
        
        # Setup the gene list
        selected_genes = node
        num_timesteps = n_timesteps
        # Check if the cluster_dict exists
        if not hasattr(self, 'cluster_dict'):
            print("Error: cluster_dict not found. Please run calculate_kmean_cluster() first.")
            return
        model_mtx = self.simulation_df.assign(type=self.simulation_df['model_id'].map(self.cluster_dict))
        cluster_type = list(model_mtx.type.unique())

        # Function to create matrix for each condition
        def create_vis_matrix(cluster_type):
            vis = model_mtx.loc[model_mtx['type'] == cluster_type]
            vis = vis[selected_genes + ['timepoint', 'model_id']]
            vis = vis[vis.timepoint.isin(range(0, num_timesteps))]
            vis['model_id'] = vis['model_id'].astype('str')
            vis['type'] = cluster_type
            return vis

        # Create matrices for each condition
        vis_matrices = [create_vis_matrix(i) for i in cluster_type]

        # Plot with Seaborn
        plt.figure(figsize=(len(cluster_type) * fig_width, len(selected_genes) * fig_height))
        n_genes = len(selected_genes)

        for i, gene in enumerate(selected_genes):
            for j, vis in enumerate(vis_matrices):
                plt.subplot(n_genes, len(cluster_type), len(cluster_type) * i + j + 1)
                plot = sns.lineplot(data=vis, x='timepoint', y=gene, lw=2, 
                                    units = 'model_id', estimator = None, alpha = 0.4)
                plot.set_ylim(0, 1.1)
                plot.set_ylabel(gene)
                plot.set_xlabel(None)
                plot.grid(True)
                if i == 0:
                    plot.set_title(f'Cluster_{cluster_type[j]}')
                    
        for i, gene in enumerate(selected_genes):
            for j, vis in enumerate(vis_matrices):
                plt.subplot(n_genes, len(cluster_type), len(cluster_type) * i + j + 1)
                plot = sns.lineplot(data=vis, x='timepoint', y=gene, lw=2, hue = 'type',palette='Set1')
                plot.set_ylim(0, 1.1)
                plot.set_ylabel(gene)
                plot.grid(True)
                plot.set_xlabel(None)
                plot.legend().remove()

        plt.tight_layout()
        if save_fig == True:
            plt.savefig('logic_trajectory.png', dpi=600, bbox_inches='tight')
            plt.savefig('logic_trajectory.pdf', bbox_inches='tight')
        plt.show()
        plt.close()


import os
import pandas as pd
from tqdm.auto import tqdm
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.cluster.hierarchy import linkage, fcluster
from sklearn.manifold import MDS

import networkx as nx

class SuccessionDiagram:
    def __init__(self, model_path, max_simulation_size = 30):
        """
        Initialize the SuccessionDiagram object.

        :param model_path: path to the .bnet files
        :param dict kwargs: parameters of the simulation
        """
        self.path = model_path
        self.max_simulation_size = max_simulation_size
        
    def calculate_succession_diagram(self):
        """
        Calculate the Succession Diagram of a Boolean network model from a given file.
        """
        try:
            from pyboolnet.file_exchange import bnet2primes
            import pystablemotifs as sm
            import pystablemotifs.export as ex

            model_path = self.path
            max_simulate_size = self.max_simulation_size
            model_files = os.listdir(self.path)
            
            # Create a dictionary object to store models
            models_net = {}
            # Loop through the models
            print('Calculating Succession Diagrams')
            for i in tqdm(model_files): 
                # Adjust model name
                model_name = i.replace('.bnet','')
                primes = bnet2primes(model_path + i)
                ar = sm.AttractorRepertoire.from_primes(primes, max_simulate_size=max_simulate_size)
                models_net[model_name]=ex.networkx_succession_diagram(ar,include_attractors_in_diagram=True)
            
            # Return the SD networks object
            self.models_net = models_net
            self.node_name = list(primes.keys())
            print('Succession Diagrams calculated')
        except ImportError:
            print("pyboolnet and pystablemotifs libraries are not installed. Please install them to calculate the succession diagram.")
            
    def calculate_sd_states(self):
        """
        Calculate the state distribution (SD) states for the given models.
        This method processes the models stored in `self.models_net` to extract and calculate
        the state distributions for each node in the models. It generates two matrices:
        `states_mtx` which contains the state values for each node, and `states_bin` which 
        contains the binary count of occurrences of each state.
        Attributes:
            states_bin (pd.DataFrame): A DataFrame containing the binary count of occurrences 
                           of each state.
            all_states (list): A list of all unique states found across the models.
        """
        # Load the model
        models_net = self.models_net
        node_name = self.node_name

        states_mtx = pd.DataFrame()
        states_bin = pd.DataFrame()
        for i in tqdm(list(models_net.keys())):
            # Extract states from GM nodes
            indexes = [data['index'] for _, data in models_net[i].nodes(data=True) if 'index' in data]
            indexes = ['A' if isinstance(i, str) else i for i in indexes]
            
            labels = [data['label'] for _, data in models_net[i].nodes(data=True) if 'label' in data]
            states = [data['states'] for _, data in models_net[i].nodes(data=True) if 'states' in data]

            # Create a new label index
            new_index = []
            for idx, index in enumerate(indexes):
                if isinstance(index, int):
                    new_index.append(labels[idx])
                elif isinstance(index, str):
                    new_index.append(index)
            new_index

            # Convert to DataFrame
            states_df = pd.DataFrame(states)
            # Convert values to int or logical value
            states_df = states_df.map(lambda x: int(x) if pd.notnull(x) else '*')

            # Set index to the matrix 
            states_df = states_df[node_name]
            states_df_strings = states_df.apply(lambda row: ''.join(row.astype(str)), axis=1)
            new_index = [new_index[j]+'_'+states_df_strings[j] for j in range(len(states_df_strings))]

            states_df.index = new_index

            # Fill NA with *
            states_df.fillna('*',inplace=True)

            # Concatenate the matrix
            states_mtx = pd.concat([states_mtx, states_df], axis=0, ignore_index = False)
            states_mtx = states_mtx[~states_mtx.index.duplicated(keep='first')]

            # Count the index from each of the models
            states_bin = pd.concat([states_bin, states_df.index.value_counts()], axis=1, ignore_index = False)
            states_bin.fillna(0, inplace=True)

        states_bin.columns = list(models_net.keys())

        # Save the states matrix and binary matrix
        self.states_bin = states_bin
        self.all_states = list(states_bin.index)

    def create_sd_networks(self):
        # Load the model
        models_net = self.models_net
        node_name = self.node_name
        all_states = self.all_states
        
        # Create network adjacency matrix
        model_adj = {}

        # For loop to create the adjacency matrix
        print('Creating SD networks')
        for i in tqdm(list(models_net.keys())):

            # Extract states from GM nodes
            indexes = [data['index'] for _, data in models_net[i].nodes(data=True) if 'index' in data]
            indexes = ['A' if isinstance(i, str) else i for i in indexes]

            labels = [data['label'] for _, data in models_net[i].nodes(data=True) if 'label' in data]
            states = [data['states'] for _, data in models_net[i].nodes(data=True) if 'states' in data]


            # Create a new label index
            new_index = []
            for idx, index in enumerate(indexes):
                if isinstance(index, int):
                    new_index.append(labels[idx])
                elif isinstance(index, str):
                    new_index.append(index)
            new_index

            # Convert to DataFrame
            states_df = pd.DataFrame(states)
            # Convert values to int or logical value
            states_df = states_df.map(lambda x: int(x) if pd.notnull(x) else '*')

            # Set index to the matrix 
            states_df = states_df[node_name]
            states_df_strings = states_df.apply(lambda row: ''.join(row.astype(str)), axis=1)
            new_index = [new_index[j]+'_'+states_df_strings[j] for j in range(len(states_df_strings))]

            # Create a matrix of the states
            model_adj[i] = nx.to_pandas_adjacency(models_net[i])
            model_adj[i].index = new_index
            model_adj[i].columns = new_index

            # Reindex the matrix according to the states_bin
            model_adj[i] = model_adj[i].reindex(all_states, axis=0).reindex(all_states, axis=1)
            model_adj[i].fillna(0, inplace=True)

        # Create a networks objects
        N = len(model_adj.keys())
        model_list = list(model_adj.keys())

        networks = []
        # Compute the distance matrix
        for i in model_list:
            G = nx.from_numpy_array(model_adj[i].to_numpy(), create_using=nx.DiGraph)
            networks.append(G)

        # Save the networks object
        self.networks = networks
        print('SD networks created')

    def calculate_sdnet_distance(self):
        
        try:
            from netrd.distance import DeltaCon
            
            networks = self.networks
            model_list = list(self.models_net.keys())
            
            # Compute DeltaCon distance matrix
            N = len(networks)
            distance_matrix = np.zeros((N, N))
            deltacon = DeltaCon()

            for i in tqdm(range(N)):
                for j in range(i+1, N):
                    dist = deltacon.dist(networks[i], networks[j])
                    distance_matrix[i, j] = dist
                    distance_matrix[j, i] = dist
            distance_matrix = pd.DataFrame(distance_matrix, index=model_list, columns=model_list)

            # Save the distance matrix
            self.distance_matrix = distance_matrix
            print('SD networks distance calculated')
        except ImportError:
            print("netrd library is not installed. Please install it to calculate the distance matrix.")
            self.distance_matrix = None

    def cluster_sdnet(self, n_cluster):
        distance_matrix = self.distance_matrix
        model_list = list(self.models_net.keys())
        
        # Compute hierarchical clustering
        Z = linkage(distance_matrix, method='ward')
        clusters = fcluster(Z, t=n_cluster, criterion='maxclust')

        # Save the clusters
        self.clusters = clusters
        print('SD networks clustered')
    
    def plot_sdnet_cluster(self):
        distance_matrix = self.distance_matrix
        clusters = self.clusters
        networks = self.networks

        # Visualize clusters with MDS
        mds = MDS(dissimilarity='precomputed', random_state=12345)
        coords = mds.fit_transform(distance_matrix)

        plt.figure(figsize=(6, 5))
        for cluster in np.unique(clusters):
            idx = np.where(clusters == cluster)
            plt.scatter(coords[idx, 0], coords[idx, 1], label=f'Cluster {cluster}')
        plt.legend()
        plt.title('Clusters Visualized via MDS')
        plt.show()

        # Plot example graphs from each cluster
        for cluster in np.unique(clusters):
            idx = np.where(clusters == cluster)[0][0]
            G = networks[idx]
            plt.figure()
            pos = nx.circular_layout(G)
            nx.draw(G, pos, with_labels=True, node_color=f"C{cluster}")
            plt.title(f'Example Network from Cluster {cluster}')
            plt.show()


import numpy as np
import pandas as pd
from tqdm.auto import tqdm

class logic_analysis:
    def __init__(self, logic_clause_flattend, ):
        self.logic = logic_clause_flattend
        

    def logic_analysis(self):
        # Get the logic rules
        logic_rules = self.logic.get_logic_rules



# General packages
import os
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

# General logic package
import mpbn

# Packages for processing logical rules
from sklearn.preprocessing import OrdinalEncoder
from scipy.stats import chi2_contingency
from scipy.cluster.hierarchy import linkage, leaves_list

# Model collections
from collections import Counter
# multiprocessing
from multiprocessing import Pool

# Define key parameters
seed = 0
np.random.seed(seed)

def clause_string(logic):
    """
    Converts a list of logical clauses into a list of string representations.

    Args:
        logic (list or bool): A list of logical clauses where each clause is a list of tuples.
                            Each tuple contains a variable and a boolean indicating its negation.
                            If the input is a boolean, it will be converted to '0' or '1'.

    Returns:
        list: A list of strings representing the logical clauses. Each clause is represented as a 
            string with variables joined by ' & '. Negated variables are prefixed with '!'.
    """
    logic_rules = []
    if logic == False:
        logic_rules = ['0']
    elif logic == True:
        logic_rules = ['1']
    else:
        for i in range(len(logic)):
            logic_clause = logic[i].copy()
            formatted_strings = [f"{t[0]}" if t[1] else f"!{t[0]}" for t in logic_clause]
            single_line_string = ' & '.join(formatted_strings)
            logic_rules.append(single_line_string)
    return(logic_rules)

def dataframe_model_dnf(model_dnf):
    """
    Converts a dictionary of Disjunctive Normal Form (DNF) logical rules into a DataFrame and 
    returns a Series with the logical expressions for each node.

    Args:
        model_dnf (dict): A dictionary where keys are node names and values are lists of clauses 
                        representing the DNF logical rules for each node.

    Returns:
        pd.Series: A Series where the index is the node names and the values are the combined 
                logical expressions in DNF format for each node.
    """
    node_names = list(model_dnf.keys())
    logic_mtx = pd.DataFrame()
    for i in node_names:
        logic_mtx = pd.concat([logic_mtx,pd.DataFrame(clause_string(model_dnf[i]))], 
                        axis = 1, ignore_index = False,)
    logic_mtx = logic_mtx.map(lambda x: f"({x})" if isinstance(x, str) else x)    
    logic_mtx=logic_mtx.transpose()
    logic_mtx = logic_mtx.fillna('')
    logic_mtx.index = node_names
    logic_full = logic_mtx.agg(' | '.join, axis=1).str.strip(' | ')
    return(logic_full)

def split_gene_clauses(model_logic_mtx, gene): 
    """
    Splits the logical clauses for a given gene in the model logic matrix.

    Args:
        model_logic_mtx (pd.DataFrame): A DataFrame containing the logical rules for various genes.
        gene (str): The gene for which the logical clauses need to be split.

    Returns:
        pd.DataFrame: A DataFrame where each row corresponds to a split clause of the original logical rules for the given gene.
    """
    split_data = [item.split(" | ") for item in model_logic_mtx.loc[gene]]
    df = pd.DataFrame(split_data, index = model_logic_mtx.loc[gene].index)
    return(df)

def logic_clause_frequency(model_logic_mtx):
    """
    Calculate the frequency of logic clauses in a given model logic matrix.

    This function processes a matrix of logical rules, splits the gene clauses,
    counts the occurrences of each clause, and computes the frequency of each
    clause across all rows. The result is a DataFrame where each column 
    represents the frequency of clauses for a corresponding index in the input matrix.

    Parameters:
    model_logic_mtx (pd.DataFrame): A DataFrame where each row represents a set of 
                                    logical rules for a model.

    Returns:
    pd.DataFrame: A DataFrame containing the frequency of each logic clause. 
                Columns correspond to the indices of the input matrix.
    """
    logic_clause = pd.DataFrame()
    for i in model_logic_mtx.index:
        df = split_gene_clauses(model_logic_mtx, i)
        df = df.fillna("0")

        from collections import Counter
        # Function to count strings in a row
        def count_strings_in_row(row):
            return Counter(row)
        # Apply the function to each row and convert the result to a DataFrame
        row_counts = df.apply(lambda row: pd.Series(Counter(row)), 
                                    axis=1).fillna(0)
        clauses = row_counts.sum()/len(df)
        logic_clause = pd.concat([logic_clause,clauses], ignore_index = True, axis = 1)
    logic_clause= logic_clause.fillna(0)
    logic_clause.columns = model_logic_mtx.index
    return(logic_clause)

# Define function to process each model file
def process_model_file(file):
    model = mpbn.load(file)
    model = model.as_dnf()
    model = dataframe_model_dnf(model)
    model.name = os.path.splitext(os.path.basename(file))[0]
    return model

# Define function to process each model's clauses
def process_model_name(i, model_logic_mtx):
    clause = logic_clause_frequency(model_logic_mtx[[i]]).transpose()
    clause = clause.stack()
    clause.index = clause.index.map('_'.join)
    clause = clause.loc[~(clause == 0)]
    clause.name = i
    return clause

class logic:
    def __init__(self, path):
        """
        Loads and processes logical models from a specified directory.

        Args:
            path (str): The directory path containing the model files.

        Returns:
            pd.DataFrame: A DataFrame containing the concatenated logical models in DNF format.

        The function performs the following steps:
        1. Lists all files in the specified directory.
        2. Initializes an empty DataFrame to store the model logic.
        3. Iterates over each file in the directory:
            a. Loads the model file.
            b. Converts the model to Disjunctive Normal Form (DNF).
            c. Converts the DNF model to a DataFrame.
            d. Sets the model name based on the file name (excluding the extension).
            e. Concatenates the model DataFrame to the main DataFrame.
        4. Attaches the model logic DataFrame to the class.
        """
        # Define models path
        # os.chdir(path)
        model_files = [os.path.join(path, model_file) for model_file in os.listdir(path)]
        model_logic = pd.DataFrame()

        print('Loading models logics')
        # Process all model files in parallel
        with Pool() as pool:
            results = list(tqdm(pool.imap(process_model_file, model_files), total=len(model_files)))

        # Concatenate results into a single DataFrame
        print('Concatenate results into matrix')
        for model in tqdm(results):
            model_logic = pd.concat([model_logic, model], axis=1, ignore_index=False)

        # Attach the model logic to the class
        self.model_logic = model_logic
        self.path = path

    def create_hash(self):
        """
        Creates a hash representation for each model file in the specified directory and stores it in the class instance.
        This method performs the following steps:
        1. Defines the model paths.
        2. Lists all model files in the specified directory.
        3. Converts each model to a hash representation using a for loop.
        4. Creates model names by removing the '.bnet' extension from each file name.
        5. Converts the list of hash models to a pandas Series.
        6. Attaches the hash models to the class instance.
        Attributes:
        -----------
        self.path : str
            The directory path where model files are located.
        self.hash_models : pandas.Series
            A Series containing the hash representations of the models, indexed by model names.
        Prints:
        -------
        'Convert model logics to hash' : str
            Indicates the start of the hash conversion process.
        'Hash of models created' : str
            Indicates the completion of the hash conversion process.
        """
        
        # Define model paths
        path = self.path

        # Define models
        model_files = os.listdir(path)

        # For loop to convert models to hash
        hash_models = []
        print('Convert model logics to hash')
        for i in tqdm(model_files):
            model = mpbn.load(path + i)
            model = model.make_hash()
            hash_models.append(model)
        
        # Create model names
        model_names = [file.replace('.bnet', '') for file in model_files]

        # Convert to numpy array
        hash_models = pd.Series(hash_models)
        hash_models.index = model_names

        # Attach the hash models to the class
        self.hash_models = hash_models
        print('Hash of models created')
    
    def count_logic_function(self):
        model_logic = self.model_logic
        model_logic_t= model_logic.transpose()

        # Ordinal encoding
        encoder = OrdinalEncoder()
        encoded_data = encoder.fit_transform(model_logic_t)
        encoded_df = pd.DataFrame(encoded_data, columns=model_logic_t.columns)

        # 
        sort_index = encoded_df.max().sort_values().index

    def create_flattend_logic_clause(self):
        """
        Flattens the logical clauses from a given model logic matrix.

        This function processes a DataFrame where each column represents a model's logical clauses.
        It computes the frequency of each clause, stacks them, and concatenates them into a single
        DataFrame. The resulting DataFrame has the same columns as the input, with rows representing
        the flattened logical clauses and their frequencies.

        Parameters:
        model_logic_mtx (pd.DataFrame): A DataFrame where each column represents a model's logical clauses.

        Returns:
        pd.DataFrame: A DataFrame with flattened logical clauses as rows and models as columns, 
                    filled with the frequency of each clause.
        """
        model_logic_mtx = self.model_logic
        model_name = list(model_logic_mtx.columns)
        logic_clause_flattend = pd.DataFrame()

        print('Flatten models logic clauses')
        # Use multiprocessing to speed up the process
        with Pool() as pool:
            results = list(tqdm(pool.starmap(process_model_name, [(i, model_logic_mtx) for i in model_name]), total=len(model_name)))
        print('Concatenate results into matrix')
        # Concatenate the results
        for clause in tqdm(results):
            logic_clause_flattend = pd.concat([logic_clause_flattend, clause], ignore_index=False, axis=1)

        # Fill NA to 0
        logic_clause_flattend = logic_clause_flattend.fillna(0)

        # Attach the model logic to the class
        self.logic_clause_flattend = logic_clause_flattend
        print('Flattend logic clause created')

    def map_model_clusters(self, cluster):

        # Map model names to their clusters
        self.cluster_dict = cluster
        model_cluster = pd.Series(cluster, name='cluster')
        model_logic = self.model_logic

        # Create a DataFrame to store the logic length for each cluster
        var_logic_clust = pd.DataFrame()
        for i in list(model_cluster.unique()):
            model_logic_sub = model_logic.transpose().loc[list(model_cluster.index[model_cluster == i])]
            #model_logic_sub = model_logic_sub.drop(['logic_cluster'], axis = 1)
            var_logic_length = []
            for j in model_logic_sub.columns:
                num_logic = model_logic_sub[j].value_counts().size
                var_logic_length.append(num_logic)
            var_logic_length = pd.DataFrame(var_logic_length, columns = ['logic_length'], index = model_logic_sub.columns)
            var_logic_clust = pd.concat([var_logic_clust, var_logic_length], axis = 1, ignore_index = False)
        var_logic_clust.columns = list(model_cluster.unique())

        # Ordinal encoding
        encoder = OrdinalEncoder()
        model_logic_t= model_logic.transpose()
        encoded_data = encoder.fit_transform(model_logic_t)
        encoded_df = pd.DataFrame(encoded_data, columns=model_logic_t.columns)

        # Sort the index based on the maximum value in each column
        self.encoded_df = encoded_df
        self.sort_index = encoded_df.max().sort_values().index
        self.var_logic_clust = var_logic_clust.loc[self.sort_index]

        print('Model clusters mapped to logic clauses')

    def calculate_logic_statistic(self, pval_threshold = 0.0001):
        """
        Calculates the statistics of logical clauses in the model logic matrix.

        This function processes the logical clauses in the model logic matrix, performs a chi-square test
        to identify marker and varied features, and creates a DataFrame summarizing the results.

        Returns:
            pd.DataFrame: A DataFrame containing the chi-square test results, feature groups, and additional
                        information about each feature.
        """
        logic_df = self.logic_clause_flattend.transpose()
        model_cluster = logic_df.index.map(self.cluster_dict)

        # Step 1: Identify constant features.
        constant_features = [col for col in logic_df.columns if logic_df[col].nunique() == 1 and col != 'group']

        # Step 2 & 3: For non-constant features, perform chi-square test.
        marker_features = []
        varied_features = []
        chi2_results = []

        for col in logic_df.columns:
            if col in constant_features or col == 'group':
                continue
            contingency_table = pd.crosstab(logic_df[col], model_cluster)
            chi2, p, _, _ = chi2_contingency(contingency_table)
            chi2_results.append({'Feature': col, 'chi2': chi2, 'p_value': p})
            if p < pval_threshold:
                marker_features.append(col)
            else:
                varied_features.append(col)
        chi2_df = pd.DataFrame(chi2_results).set_index('Feature')

        # Create a dictionary to store features and their groups
        feature_groups = {
            'Feature': constant_features + varied_features + marker_features,
            'Group': (['Constant'] * len(constant_features)) +
                    (['Varied'] * len(varied_features)) +
                    (['Marker'] * len(marker_features))
        }
        features_df = pd.DataFrame(feature_groups).set_index('Feature')

        # Combine two DataFrames
        stat_logic_df = pd.concat([chi2_df, features_df], axis=1, ignore_index=False)
        stat_logic_df['chi2'] = stat_logic_df['chi2'].replace({np.nan: 0})
        stat_logic_df['p_value'] = stat_logic_df['p_value'].replace({np.nan: 1})

        # Extract Node and Regulation from the index
        stat_logic_df['Node'] = [col.split('_')[0] if '_' in col else col for col in stat_logic_df.index]
        stat_logic_df['Regulation'] = [col.split('_')[1] if '_' in col else col for col in stat_logic_df.index]

        # Add back to the model.logic object
        self.stat_logic_df = stat_logic_df
        self.pval_threshold = pval_threshold

    def plot_manhattan(self, fig_size=(10, 5), show_label = False, save_fig = False):

        plt.figure(figsize=fig_size)
        stat_logic_df = self.stat_logic_df
        stat_logic_df['-log10_p'] = -np.log10(self.stat_logic_df['p_value'])

        # Add jitter to x positions
        nodes = stat_logic_df['Node'].astype('category')
        x = nodes.cat.codes + np.random.uniform(-0.2, 0.2, size=len(stat_logic_df))

        group_palette = {'Constant': '#bdbdbd', 'Varied': '#1f77b4', 'Marker': '#d62728'}
        colors = stat_logic_df['Group'].map(group_palette)
        ax = plt.scatter(x, stat_logic_df['-log10_p'], 
                        s=200, alpha=0.7, 
                        linewidths=1, edgecolor='black', c=colors)

        if show_label == True:
            # Annotate points with feature names
            for i, row in stat_logic_df.iterrows():
                plt.annotate(row.name, (x[i], row['-log10_p']),
                            textcoords="offset points", xytext=(0,5), ha='center', fontsize=9, color='black')
        
        plt.axhline(-np.log10(self.pval_threshold), color='red', linestyle='--', label=f'p={self.pval_threshold}')
        plt.ylabel('-log10(p-value)')
        plt.xlabel('Node')
        plt.title('Manhattan Plot of Chi-square p-values by Node')
        plt.xticks(ticks=range(len(nodes.cat.categories)), labels=nodes.cat.categories, rotation=90)
        plt.legend()
        plt.tight_layout()
        if save_fig == True:
            plt.savefig('manhattan_plot.png', dpi=600, bbox_inches='tight')
            plt.savefig('manhattan_plot.pdf', bbox_inches='tight')
        plt.show()
        plt.close()

    def plot_logicstat_summary(self, fig_size=(12, 7), save_fig = False):

        encoded_df = self.encoded_df
        sorted_var_logic_length = encoded_df.max().sort_values() + 1
        # Create a figure with two subplots sharing the x-axis
        fig, axes = plt.subplots(2, 1, figsize=fig_size, sharex=True, gridspec_kw={'height_ratios': [2, 1]})

        # Plot 1
        ## Plot the first barplot
        sns.barplot(
            x=sorted_var_logic_length.index, 
            y=sorted_var_logic_length.values, 
            hue=sorted_var_logic_length.values,
            palette=sns.color_palette("viridis", as_cmap=True), 
            edgecolor='black', linewidth=0.8, ax=axes[0], width=0.8
        )
        ## Annotate the total number of logics
        for index, value in enumerate(sorted_var_logic_length.astype('int')):
            axes[0].text(index, value + 0.1, str(value), ha='center', va='bottom', fontsize=15)
        ## Customize the first plot
        axes[0].axhline(0, color="k", clip_on=False)
        axes[0].set_ylabel('Number of Logics', fontsize=15)
        axes[0].set_title('Number of Possible Logics and Clauses', fontsize=16)
        axes[0].tick_params(axis='x', rotation=90, labelsize=15)
        axes[0].tick_params(axis='y', labelsize=15)
        axes[0].set_ylim(0, sorted_var_logic_length.max() + 1)
        axes[0].grid(linestyle='--', linewidth=0.5)
        axes[0].set_axisbelow(True)

        # Plot 2
        ## Plot the second stacked barplot
        features_df_grouped = self.stat_logic_df.groupby(['Node', 'Group']).size().unstack(fill_value=0)
        features_df_grouped = features_df_grouped.loc[sorted_var_logic_length.index]
        features_df_grouped.plot(
            kind='bar', stacked=True, colormap='tab10', ax=axes[1], edgecolor='black', linewidth=0.8, width=0.8
        )
        ## Annotate the total number of clauses
        total_clauses = features_df_grouped.sum(axis=1)
        for index, value in enumerate(total_clauses.astype('int')):
            axes[1].text(index, value + 0.1, str(value), ha='center', va='bottom', fontsize=15)
        ## Customize the second plot
        axes[1].set_ylabel('Number of Clauses', fontsize=15)
        axes[1].set_xlabel(None)
        axes[1].legend(title='Group', bbox_to_anchor=(1.0, 1), loc='upper left', fontsize=15, title_fontsize=15)
        axes[1].tick_params(axis='x', rotation=90, labelsize=15)
        axes[1].tick_params(axis='y', labelsize=15)
        axes[1].set_ylim(0, total_clauses.values.max() + 1)
        axes[1].grid(linestyle='--', linewidth=0.5)
        axes[1].set_axisbelow(True)

        # Ensure both subplots have the same x-ticks and labels
        xticks = range(len(sorted_var_logic_length.index))
        axes[1].set_xticks(xticks)
        axes[1].set_xticklabels(sorted_var_logic_length.index, rotation=90, fontsize=15)
        axes[0].set_xticks(xticks)
        axes[0].set_xticklabels(sorted_var_logic_length.index, rotation=90, fontsize=15)

        plt.tight_layout()
        if save_fig == True:
            plt.savefig('logicstat_summary.png', dpi=600, bbox_inches='tight')
            plt.savefig('logicstat_summary.pdf', bbox_inches='tight')
        plt.show()
        plt.close()

    def plot_node_logic_heatmap(self, 
                                node, 
                                fig_size=(10, 8), save_fig = False):       
        
        # Load the logic statistics DataFrame 
        stat_logic_df = self.stat_logic_df
        selected_features = stat_logic_df.loc[(stat_logic_df['Node'].isin(node))].index

        # Create a color palette for the groups
        test = self.logic_clause_flattend
        test = test.transpose()
        test['group'] = test.index.map(self.cluster_dict)

        unique_groups = test['group'].unique()
        palette = sns.color_palette("tab10", n_colors=len(unique_groups))
        group_colors = test['group'].map(dict(zip(unique_groups, palette)))

        # Create the clustermap with row colors and separated cluster groups
        # First, sort the DataFrame by cluster group to visually separate them
        sorted_idx = test.sort_values('group').index

        # Get the split point between the two clusters
        split_point = (test.loc[sorted_idx, 'group'] == unique_groups[1]).idxmax()

        # Perform row clustering only within each cluster group
        for i in range(len(unique_groups)):
            group_idx = test.loc[sorted_idx, 'group'] == unique_groups[i]
            linkage_matrix = linkage(test.loc[sorted_idx[group_idx], selected_features], method='average', metric='euclidean')
            leaves = sorted_idx[group_idx][leaves_list(linkage_matrix)]
            if i == 0:
                clustered_idx = leaves
            else:
                clustered_idx = clustered_idx.append(leaves)

        # Create the clustermap
        plt.figure(figsize=fig_size)
        sns.clustermap(
            test.loc[clustered_idx, selected_features].transpose(),
            col_cluster=False,
            row_cluster=False,
            cmap='viridis',
            figsize=(10, 5),
            col_colors=group_colors.loc[clustered_idx]
        )

        if save_fig == True:
            plt.savefig('node_logic_heatmap.png', dpi=600, bbox_inches='tight')
            plt.savefig('node_logic_heatmap.pdf', bbox_inches='tight')

        plt.show()
        plt.close()


"""
{Test script for the Boolean Benchmark project}
"""
__author__ = 'Saran PANKAEW'
__version__ = '0.4'
__maintainer__ = 'Saran PANKAEW'
__email__ = 'saran.pankeaw@curie.fr'
__status__ = 'development'
__date__ = '05/08/2025'

from .model_attractors import attractors
from .model_simulation import simulation
from .trajectory_clustering import trajectory
from .logical_rules_processing import logic
from .succession_diagram import SuccessionDiagram
# from .logic_analysis import analysis

class LogicEnsemble:
    def __init__(self, path, project_name):
        self.path = path
        self.project = project_name

    def create_attractor(self):
        self.attractor = attractors(self.path)
        print('Attractor object created')    

    def create_simulation(self):
        self.simulation = simulation(self.path)
        print('Simulation object created')

    def create_trajectory(self, parameters = None):
        # Check if the simulation object is created
        if self.simulation is None:
            print('Simulation object is not created yet')
            return
        else:
            self.trajectory = trajectory(self.simulation.simulation_df)
            print('Trajectory object created')

    def create_logic(self):
        self.logic = logic(self.path)
        print('Logic object created')

    def create_succession_diagram(self):
        self.succession_diagram = SuccessionDiagram(self.path)
        print('Succession diagram object created')
    
    def __repr__(self):
        print(f'LogicEnsemble object for the project {self.project}')
        print(f'Path: {self.path}')
        print(f'Simulation object: {self.simulation}')

    
    


import boolsim
from colomoto.minibn import BooleanNetwork
from colomoto_jupyter import tabulate # for display
import os
import pandas as pd
from multiprocessing import Pool
from tqdm.auto import tqdm
import gc

"""
This script is used to generate sequence of attractors from the model.
"""

def calculate_attractors(file):
        """
        Calculate the attractors of a Boolean network model from a given file.
        Args:
            file (str): The path to the file containing the Boolean network model.
        Returns:
            pandas.DataFrame: A DataFrame where each column represents an attractor 
                              in binary string format, and the index is the name of 
                              the file (without the '.bnet' extension).
        """

        # Load the model
        bn = BooleanNetwork.load(file)
        a = boolsim.attractors(bn, update_mode='asynchronous')
        
        # Make sure a has all the columns based on the model
        for i in range(len(a)):
            for key in bn.keys():
                if key not in a[i]:
                    a[i][key] = 0
            a[i] = {k: a[i][k] for k in bn.keys()}

        # Convert each dictionary in the list to a string of binary values
        binary_strings = [''.join(str(value) for value in attractor.values()) for attractor in a]
        attractors = pd.DataFrame([binary_strings],columns=binary_strings)
        attractors[:]=1
        attractors.index = [file.replace('.bnet', '')]

        # Return attractors list
        return attractors.T

class attractors:
    def __init__(self, model_path):
        """
        Initializes the ModelAttractors class with the given model path.

        Args:
            model_path (str): The file path to the Boolean network model.

        Attributes:
            path (str): The file path to the Boolean network model.
            attractors_df (pandas.DataFrame or None): DataFrame to store attractors, initialized as None.
        """
        self.path = model_path
        self.attractors_df = None
    
    def get_attractors(self, num_cores=10, chunksize=10):
        """
        Improved multiprocessing version with better memory management.
        """
        # os.chdir(self.path)
        model_files = [os.path.join(self.path, f) for f in os.listdir(self.path) if f.endswith('.bnet')]
        model_logic = pd.DataFrame()
        
        # Use imap with chunking for better memory efficiency
        with Pool(processes=num_cores) as pool:
            # Process in chunks to avoid memory buildup
            for result in tqdm(
                pool.imap(calculate_attractors, model_files, chunksize=chunksize), 
                total=len(model_files),
                desc="Processing models"
            ):
                model_logic = pd.concat([model_logic, result], axis=1, ignore_index=False)
                # Periodic garbage collection
                if len(model_logic.columns) % 100 == 0:
                    gc.collect()
        
        self.attractors_df = model_logic.fillna(0)
        print('Attractors calculation completed')



import maboss
from tqdm.auto import tqdm
import collections
import os
import pandas as pd

"""
This script is used to simulate the model with the given parameter set.
"""
_default_parameter_list = collections.OrderedDict([
    ('time_tick', 1),
    ('max_time', 100),
    ('sample_count', 1000),
    ('discrete_time', 0),
    ('use_physrandgen', 1),
    ('seed_pseudorandom', 0),
    ('display_traj', 0),
    ('statdist_traj_count', 0),
    ('statdist_cluster_threshold', 1),
    ('thread_count', 1),
    ('statdist_similarity_cache_max_size', 20000)
])

class simulation:
    """
    Attributes:
        network (Network): A Network object, that will be translated in a bnd file.
        mutations (list): A list of nodes for which mutation can be triggered by modifying the cfg file.
        palette (dict): A mapping of nodes to color for plotting the results of the simulation.
        param (dict): A dictionary that contains global variables (keys starting with a '$'), and simulation parameters (keys not starting with a '$').
    Methods:
        __init__(model_path, parameters=collections.OrderedDict({}), **kwargs):
                model_path (str): Path to the .bnet files.
                parameters (dict, optional): Parameters of the simulation. Defaults to an empty OrderedDict.
                kwargs (dict): Additional parameters of the simulation.
        update_parameters(**kwargs):
            Add elements to `self.param`.
                **kwargs: Arbitrary keyword arguments to be added to `self.param`.
        mutate(node, value):
                node (str): The name of the node to be mutated.
                value (int): The value of the mutation (0 or 1).
        run_simulation(output_nodes=None):
    """
    def __init__(self, model_path, parameters=collections.OrderedDict({}), **kwargs):
        """
        Initialize the Simulation object.

        :param model_path: path to the .bnet files
        :param dict kwargs: parameters of the simulation
        """
        self.path = model_path
        self.param = _default_parameter_list.copy()
        self.palette = {}
        self.mutations = {}
        self.mutationTypes = {}
        self.refstate = {}
        
    def update_parameters(self, **kwargs):
        """
        Add elements to ``self.param``.

        This method updates the parameters stored in the `self.param` dictionary.
        It accepts keyword arguments and adds them to `self.param` if they are 
        present in the `_default_parameter_list` or if their key starts with a '$'.
        If a parameter is not recognized, a warning message is printed.

        Parameters:
        **kwargs: Arbitrary keyword arguments.
            The keyword arguments to be added to `self.param`.

        Example:
        >>> obj.update_parameters(param1=value1, param2=value2)
        """

        for p in kwargs:
            if p in _default_parameter_list or p[0] == '$':
                self.param[p] = kwargs[p]
            else:
                print("Warning: unused parameter %s" % p)

    def mutate(self, condition ,node, value):
        """
        Add a mutation to the simulation.

        This method adds a mutation to the simulation. It accepts a node name and a value
        (0 or 1) as arguments. The mutation is stored in the `self.mutations` list.

        Parameters:
        node: str
            The name of the node to be mutated.
        value: int
            The value of the mutation (0 or 1).

        Example:
        >>> obj.mutate("node1", 0)
        """

        self.mutations[condition] = (node,value)

    def run_simulation(self, output_nodes = None, initial_state = None, mutation = None):
        """
        Run simulations for a list of models and store the results.
        Parameters:
        output_nodes (list, optional): List of nodes to set as output for the simulation. 
                                       If None, all nodes will be set as output. Default is None.
        Returns:
        None: The results of the simulations are stored in the `self.simulation_df` attribute.
        The function performs the following steps:
        1. Initializes an empty dictionary to store the results of each model.
        2. Retrieves the list of models from the specified model path.
        3. Iterates over each model in the list:
           a. Loads the model using `maboss.loadBNet`.
           b. Updates the model parameters with `self.param`.
           c. Sets the output nodes for the simulation.
           d. Runs the simulation.
           e. Retrieves the probability trajectory matrix of the nodes.
           f. Adds model ID and timepoint information to the matrix.
           g. Stores the matrix in the results dictionary.
        4. Concatenates the results from all models into a single DataFrame.
        5. Saves the concatenated DataFrame to the `self.simulation_df` attribute.
        """
        
        # Simulation results object
        ensemble_results = {}
        path = self.path
        model_list = os.listdir(path)
        print('Start simulation')
        
        # For loop to run the simulation
        for model in tqdm(model_list):
            # Load model
            simulations = maboss.loadBNet(path + model)

            # Setup the model initial condition
            simulations.param = self.param

            # Set the initial condition
            if initial_state is not None:
                node_names = simulations.network.names
                assigned_node = list(initial_state.keys())
                unassigned_node = list(set(node_names) - set(assigned_node))

                # Set the initial condition - assigned node
                for i in assigned_node:
                    simulations.network.set_istate(i, [1 - initial_state[i], initial_state[i]])

                # Set the initial condition - unassigned node
                for i in unassigned_node:
                    simulations.network.set_istate(i, [0.5, 0.5])
            
            # Set the mutation condition
            if mutation is not None:
                condition = self.mutations[mutation]
                # Set the condition
                simulations.mutate(condition[0],condition[1])

            # Set the output of the simulation
            if output_nodes is not None:
                simulations.network.set_output(output_nodes)
            else:
                simulations.network.set_output(simulations.network.names)

            # Perform simulations
            result = simulations.run()
            # Get matrix
            model_mtx = result.get_nodes_probtraj().copy()

            # Setup cell matrix
            ## Cells
            model_mtx['model_id'] = model.replace('.bnet','')
            ## Timepoint
            model_mtx['timepoint'] = model_mtx.index
            ## Change index
            model_mtx.index = model_mtx.index.map(str)
            model_mtx.index = model + '_' + model_mtx.index

            # Concatenate model results in dictionary
            ensemble_results[model] = model_mtx

        # Save the simulation to /tmp folder
        simulation_df = pd.concat(ensemble_results.values(), ignore_index = True)
        
        # Save the simulation to the object
        self.simulation_df = simulation_df
        print('Simulation completed')

    def run_states_simulation(self, output_nodes = None, initial_state = None, mutation = None):
        # Simulation results object
        ensemble_results = {}
        path = self.path
        model_list = os.listdir(path)
        print('Start simulation')
        
        # For loop to run the simulation
        for model in tqdm(model_list):
            # Load model
            simulations = maboss.loadBNet(path + model)
            
            # Setup the model initial condition
            simulations.param = self.param

            # Set the initial condition
            if initial_state is not None:
                node_names = simulations.network.names
                assigned_node = list(initial_state.keys())
                unassigned_node = list(set(node_names) - set(assigned_node))

                # Set the initial condition - assigned node
                for i in assigned_node:
                    simulations.network.set_istate(i, [1 - initial_state[i], initial_state[i]])

                # Set the initial condition - unassigned node
                for i in unassigned_node:
                    simulations.network.set_istate(i, [0.5, 0.5])
            
            # Set the mutation condition
            if mutation is not None:
                condition = self.mutations[mutation]
                # Set the condition
                simulations.mutate(condition[0],condition[1])

            # Set the output of the simulation
            if output_nodes is not None:
                simulations.network.set_output(output_nodes)
            else:
                simulations.network.set_output(simulations.network.names)

            # Perform simulations
            result = simulations.run()
            # Get matrix
            model_mtx = result.get_last_states_probtraj().copy()

            # Setup cell matrix
            ## Cells
            model_mtx['model_id'] = model.replace('.bnet','')
            ## Timepoint
            model_mtx['timepoint'] = model_mtx.index
            ## Change index
            model_mtx.index = model_mtx.index.map(str)
            model_mtx.index = model + '_' + model_mtx.index

            # Concatenate model results in dictionary
            ensemble_results[model] = model_mtx

        # Save the simulation to /tmp folder
        states_df = pd.concat(ensemble_results.values(), ignore_index = True)
        states_df.fillna(0, inplace=True
                         )
        # Save the simulation to the object
        self.states_df = states_df
        print('Simulation completed : object states_df has been created')